
(* general purpose routines *)

structure Gen = struct 
local open Types ;
in

fun upto (m, n) = if (m > n) then [] else m :: upto (m+1, n) ;

(* filtrev : ('a -> bool) -> 'a list -> 'a list -> 'a list
  filter and reverse list onto acc *)
fun filtrev p [] acc = acc
  | filtrev p (x :: xs) acc = filtrev p xs (if p x then x :: acc else acc) ; 

(* maprev : ('a -> 'b) -> 'a list -> 'b list -> 'b list
  map by f and reverse list onto acc *)
fun maprev f [] acc = acc
  | maprev f (x :: xs) acc = maprev f xs (f x :: acc) ;
  
(* lexComp : ('a * 'b -> order) -> 'a list * 'b list -> order
  lexicographic ordering of lists *)
fun lexComp comp ([], []) = EQUAL
  | lexComp comp ([], _) = LESS 
  | lexComp comp (_, []) = GREATER 
  | lexComp comp (h1 :: t1, h2 :: t2) = 
    let val c = comp (h1, h2) 
    in if c = EQUAL then lexComp comp (t1, t2) else c end ;

(* trComp : ('a -> 'b) -> ('b * 'b -> 'order) -> 'a * 'a -> 'order
  transforms items using tr, then compares them using comp *)
fun trComp tr comp (a, b) = comp (tr a, tr b) ;

(* group : ('a * 'a -> bool) -> 'a list -> 'a list list
  groupeq : ('a -> ''b) -> 'a list -> 'a list list
  group members of input list - members go in same group 
  when consecutive members related by f
  groupeq - consecutive members of input list grouped if f gives same result 
  grouprev - tail recursive version, group f (rev l) = grouprev f l [] 
  *)
fun group f [] = []
  | group f [x] = [[x]] 
  | group f (x1 :: x2 :: xs) = 
    let val xgs as xg :: xgs' = group f (x2 :: xs) ;
    in if f (x1, x2) then (x1 :: xg) :: xgs' else [x1] :: xgs end ;

fun groupeq f = group (fn (a, b) => f a = f b) ;

fun grouprev f [] acc = acc
  | grouprev f (x :: xs) (acc as (y :: ys) :: yss) = 
    if f (x, y) then grouprev f xs ((x :: y :: ys) :: yss)
    else grouprev f xs ([x] :: acc) 
  | grouprev f (x :: xs) [] = grouprev f xs [[x]] ;

fun groupreq f l = grouprev (fn (a, b) => f a = f b) l [] ;

(* findBt : ('a * 'a -> order) -> 'a -> 'a btree -> 'a option
  find entry in binary search tree *)
fun findBt comp str BtEmpty = NONE
  | findBt comp str (Btree (left, mid, right)) =
    case comp (str, mid) of
        LESS => findBt comp str left
      | EQUAL => SOME mid
      | GREATER => findBt comp str right ;

(* mkBt : int -> 'a list -> 'a btree * 'a list
  make a binary search tree, from first n items of an ordered list,
  returns tree and rest of list,
  usual top call is val (outbt, [] = mkBt (length list) list *)
fun mkBt n [] = (BtEmpty, [])
  | mkBt 0 list = (BtEmpty, list)
  | mkBt n list =
    let val (left_tree, mid :: right_elts) = mkBt (n div 2) list ;
      val (right_tree, rest) = mkBt ((n-1) div 2) right_elts ;
    in (Btree (left_tree, mid, right_tree), rest) end ;

(* takeWhile : ('a -> bool) -> 'a list -> 'a list
  takes elements from list as long as they all satisfy p *)
fun takeWhile p [] = []
  | takeWhile p (e :: es) = 
    if p e then e :: takeWhile p es else [] ;

(* dropUntil : ('a -> bool) -> 'a list -> 'a list 
  drops entries not satisfying p off front of list *)
fun dropUntil p (e :: es) =
    if p e then e :: es else dropUntil p es
  | dropUntil p [] = [] ;

(* split : ('a -> bool) -> 'a list -> 'a list * 'a list
  splits list, left part is longest prefix all satisfying p *)
fun split p [] = ([], []) 
  | split p (list as e :: es) = 
    if p e then let val (rh, rt) = split p es in (e :: rh, rt) end 
    else ([], list) ;

(* lfields : ('a -> bool) -> 'a list -> 'a list list
  group items into lists, separated by one or more instances of lines not
  satisfying p *)
fun lfields p l = 
  let 
    fun lfields' [] = []
      | lfields' l = 
	let val (tk, dp) = split p l ;
	in tk :: lfields' (dropUntil p dp) end ;
  in lfields' (dropUntil p l) end ;

(* takeDrop : int -> 'a list -> 'a list * 'a list
  splits list into two, with n (if there are that many) in first part *) 
fun takeDrop n [] = ([], []) (* ok if list too short *)
  | takeDrop 0 l = ([], l)
  | takeDrop n (h :: t) = 
    let val (tk, dp) = takeDrop (n-1) t in (h :: tk, dp) end ;

(* replicate : 'a -> 'a list -> 'a list
  replicate item to make list *)
fun replicate _ 0 = []
  | replicate x n = x :: replicate x (n-1) ;

(* separate : 'a -> 'a list -> 'a list
  insert separator between items in list *)
fun separate s [] = []
  | separate s [e] = [e]
  | separate s (e :: es) = e :: s :: separate s es ;

(* fmtil : int list -> string, format list of integers *)
fun fmtil il = String.concat (separate "," (map Int.toString il)) ;

fun printEvent (ExclDist (countDetails as (exclcand, tv, cts))) = 
      ("votes of candidate " ^ Int.toString exclcand ^ " of value " ^ 
      tvprint tv ^ ", from counts " ^ fmtil cts ^ "\n") 
  | printEvent (Surplus (cand, newtv, cnt)) = 
      ("surplus votes of elected candidate " ^ Int.toString cand ^
      ", from count " ^ Int.toString cnt ^ "\n") 
  | printEvent (InitDecd decd) = 
    ("initial count" ^ 
      (case decd  of [] => "\n"
	| decd => ", candidate(s) " ^ fmtil decd ^ " deceased\n")) 
  | printEvent (Elect []) = ""
  | printEvent (Elect cs) = ("candidates " ^ fmtil cs ^ " elected\n") 
  | printEvent (ExclTie {count = bct, lows, excl, votes}) =
      ("tie in determining candidate to be excluded, candidates " ^
      fmtil lows ^ " with " ^ Int.toString votes ^ 
      " votes,\n  going back to count " ^ Int.toString bct ^ 
      ", result " ^ Int.toString excl ^ " excluded\n")
  | printEvent (SurpTie {count = bct, order, votes}) =
      ("tie in elected candidates with " ^ Int.toString votes ^
      " votes,\n  going back to count " ^ Int.toString bct ^ 
      ", result " ^ fmtil order ^ "\n")
  | printEvent (Cl_1_2 loss) = 
    Int.toString loss ^ " votes lost due to clause 1(2)\n"
  | printEvent (CommLot (cands, succ)) =
    "commissioner's lot used, candidates " ^ fmtil cands ^ 
      "; " ^ Int.toString succ ^ " chosen\n" 
  | printEvent (ExclDet cand) =
    "Excluding candidate " ^ Int.toString cand ^ "\n"
  | printEvent (SurpZero {cand, count}) =
    "successful candidate " ^ Int.toString cand ^ ", elected at count " ^
    Int.toString count ^ 
    ",\n  had zero surplus (would otherwise be counted next count)\n"
  | printEvent (FinCont cands) =
    "only candidates " ^ fmtil cands ^ " remaining, all elected\n"
  | printEvent event = "" ;

(* eventStr : int * elecEvent -> string *)
fun eventStr (cnt, event) = 
  let val eventstr = printEvent event ;
  in if eventstr = "" then ""
    else ("Count " ^ Int.toString cnt ^ ": " ^ eventstr) end ;

(* recordEvent : info -> elecEvent -> unit
  records a new event, pushing event and current count onto events list,
  except Elect [] *)
fun recordEvent (info as Info {count, events, ...}) (Elect []) = () 
  | recordEvent (info as Info {count, events, ...}) event = 
    let val eventstr = eventStr (!count, event) ;
    in push (!count, event) events ;
      print eventstr end ;

fun printElec (info as Info {events = ref evs, ...}) =
  let val elecEvents = rev (List.filter (fn (_, ev) => elec ev <> []) evs) ;
    val eestrings = map eventStr elecEvents ;
  in print "SUMMARY OF ELECTED CANDIDATES\n" ;
    app print eestrings end ;
  
end ; (* local open Types ; *)
end ; (* structure Gen *)

