
(** votes in string format - internally a string of bytes,
  externally a string of bytes preceded by one byte to indicate length **)

structure RPV = struct 

(* printsvote : outstream -> string -> unit 
  printsvotes : string list -> string -> unit
  print votes in string format *)
fun printsvote ostr vote = (
  BinIO.output1 (ostr, Word8.fromInt (size vote)) ; 
  BinIO.output (ostr, Byte.stringToBytes vote)) ; 

fun printsvotes votes file = 
  let val ostr = BinIO.openOut file ;
    val () = app (printsvote ostr) votes ;
    val () = BinIO.closeOut ostr ;
  in () end ;

(* readsvote : instream -> string option
  read votes in string format *)
fun readsvote istr = 
  case BinIO.input1 istr of NONE => NONE 
    | SOME lenw => 
      SOME (Byte.bytesToString (BinIO.inputN (istr, Word8.toInt lenw))) ;

(* readsvotes : string -> int -> int * string list
  read up to nv votes, also return number of votes read,
  if n = ~1, reads all votes *)
fun readsvotes file n =
  let val istr = BinIO.openIn file ;
    fun rv 0 = (0, [])
      | rv n = 
	case readsvote istr of NONE => (0, [])
	  | SOME vote => 
	    let val (nv, vs) = rv (n-1) in (nv + 1, vote :: vs) end ;
  in rv n before BinIO.closeIn istr end ;

(* this version is tail recursive, and works where the other one fails
  in PolyML due to lack of swap space *)
fun readsvotes file nballots =
  let val instr = BinIO.openIn file ;
    fun rvs 0 sf = sf
      | rvs n (sf as (nsf, vsf)) = 
	case readsvote instr of NONE => sf
	  | SOME vote => rvs (n-1) (nsf + 1, vote :: vsf) ;
    val (nfv, rev_vs) = (rvs nballots (0, [])) ;
    val () = BinIO.closeIn instr ;
  in (nfv, rev rev_vs) end ;

(* s2ilvote' : substring -> int list
  s2ilvote : string -> int list
  il2svote : int list -> string
  convert vote between string or substring and list of preferred candidates *)
fun s2ilvote' str = 
  case Substring.getc str of NONE => []
    | SOME (c, str') => ord c :: s2ilvote' str' ;

fun s2ilvote str = s2ilvote' (Substring.all str) ;
fun s2ilvote str = map ord (explode str) ;
fun il2svote il = implode (map chr il) ;

(** votes in integer list format - OBSOLETE
  internally a list of integers, externally a list of integers in decimal,
  separated by spaces, one vote per line **)

fun printvote ostr vote = (
  app (fn i => (TextIO.output (ostr, Int.toString i ^ " "))) vote ;
  TextIO.output (ostr, "\n")) ;

fun printvotes votes file = 
  let val ostr = TextIO.openOut file ;
    val () = app (printvote ostr) votes ;
    val () = TextIO.closeOut ostr ;
  in () end ;

fun readvote string = 
  let val toks = String.tokens Char.isSpace string ;
    val ints = map (Option.valOf o Int.fromString) toks ;
  in ints end ;

fun readvotes file n =
  let val istr = TextIO.openIn file ;
    fun rv 0 = []
      | rv n = 
	case TextIO.inputLine istr of "" => []
	  | line => readvote line :: rv (n-1) ;
  in rv n end ;

(* tos : string -> string -> int -> unit
  translate n votes from printable integers to byte strings *)
fun tos file sfile n =
  let val istr = TextIO.openIn file ;
    val ostr = BinIO.openOut sfile ;
    fun rv 0 = ()
      | rv n = 
	(case TextIO.inputLine istr of "" => ()
	  | line =>
	    let val str = (implode (map chr (readvote line))) ;
	    in (printsvote ostr str) end ;
	rv (n-1)) ;
  in rv n before (TextIO.closeIn istr ; BinIO.closeOut ostr) end ;

end ; (* structure RPV *)

