
structure Types = struct 

exception Finished ;

(* binary search tree *)
datatype 'a btree = Btree of 'a btree * 'a * 'a btree
               | BtEmpty ;

type cand = int ;
type count = int ;

(* type vote : original vote is base string of substring,
  substring is current remaining preferences (not including
  candidate for whom vote currently counts *)
datatype vote = Vote of substring ;

datatype tv = TV of int * int ; (* transfer value *)

fun tvcompare (TV (n1, d1), TV (n2, d2)) = Int.compare (n1 * d2, n2 * d1) ;

fun tvprint (TV (1, 1)) = "1" 
  | tvprint (TV (n, d)) = Int.toString n ^ "/" ^ Int.toString d ;

datatype group = Group of {
  cand : cand, (* vote currently count for this candidate *)
  count : int, (* votes counted at this count *)
  tv : tv, (* current value of each vote *)
  num : int, (* number of actual votes in this group *)
  tvnum : int, (* number of votes as credited to candidate *)
  votes : vote list} ; (* list of votes *)

fun cOfG (Group {cand, ...}) = cand ;
fun ctOfG (Group {count, ...}) = count ;

(* tvof : group list -> tv
  gets tv of head of list of groups (which will normally have same tv *)
fun tvof ((Group {tv, ...} : group) :: _) = tv 
  | tvof [] = raise Fail "tvof : []" ;

(* elecEvent : associated with a particular count, as follows *)
datatype elecEvent = 
    (* initial count (count 1), with listed candidates deceased,
      this is always the first entry of events list *)
    InitDecd of cand list

    (* as a result of this count, determined to exclude candidate cand *)
  | ExclDet of cand 

    (* count distributes excluded candidate's votes of this value,
      previously counted at listed counts *)
  | ExclDist of cand * tv * int list

    (* at this count, determined that these candidates elected *) 
  | Elect of cand list 

    (* distributing candidate's surplus, new value tv, 
      previously counted at given count *)
  | Surplus of cand * tv * int 
  | Cl_1_2 of int (* clause 1(2) loss of vote value *)

    (* where candidates #lows tie with lowest number #votes of votes, 
      get lowest #excl by referring back to count #count;
      note - this event occurs at the count whose results are used in choosing
      candidate to be excluded next, ie the count previous to the counts
      at which the excluded candidate's votes are distributed *)
  | ExclTie of {lows : cand list, count : int, excl : cand, votes : int} 

    (* there was a tie, and the tiebreaker procedure was used
      to determine the order of the candidates elected at this count;
      where elected candidates tie (each with #votes votes), ie contemporary
      candidates (cl 7(3)(c)), get order by referring back to count #count *)
  | SurpTie of {order : cand list, count : int, votes : int} 

    (* at next count, we would distribute this candidate's surplus,
      except for the fact that he/she got a quota exactly (at count #count) *)
  | SurpZero of {cand : cand, count : int} 
  | CommLot of cand list * cand 

    (* after this count, only right number of candidates remaining, 
      so all elected *)
  | FinCont of cand list 
  | Text of string ;

(* elec : elecEvent -> cand list, identify events electing candidates *)
fun elec (FinCont cs) = cs
  | elec (Elect (c :: cs))  = c :: cs 
  | elec ev = [] ;

(* inter : elecEvent -> bool, identify particularly interesting events *)
fun inter (Cl_1_2 _) = true
  | inter (ExclTie _) = true
  | inter (SurpTie _) = true
  | inter (SurpZero _) = true
  | inter (CommLot _) = true
  | inter _ = false ;

type resrec = {new : int * (cand * int) list, 
  count : int * (cand * int) list, 
  tot : int * (cand * int) list,
  tv : tv} ;

(* results : results of a count, as follows -
  New is change (in vote value) at this count,
  Tot is total votes for candidates after this count,
  Count is actual votes counted;
  CRes is all three
  in each case, total, then list of (cand, num votes) pairs *)

datatype results = Cres of resrec ;

(* OBSOLETE 
datatype results = New of int * (cand * int) list 
	 | Count of int * (cand * int) list 
	 | Tot of int * (cand * int) list ;

fun isCount (Count _) = true | isCount _ = false ;
fun isNew (New _) = true | isNew _ = false ;
fun isTot (Tot _) = true | isTot _ = false ;
*)

(* type status : for candidates still in race, status, including if relevant
  vote groups grouped according to current value, in descending order *)
datatype status = Elected (* and surplus distributed *)
	| ElND of (group list list) (* elected but surplus not dist'd *) 
	| Excluded 
	| Deceased
	| Cont of (group list list) ; (* continuing *)

fun statusString Elected = "Elected"
  | statusString Excluded = "Excluded"
  | statusString (Cont _)  = "Continuing"
  | statusString (ElND _)  = "El_Not_Dist" ;

(* elec : info about newly elected candidate, 
  (candidate, number of votes, 
  group of votes which made him/her successful,
  count at which successful) *)
datatype elec = Elec of cand * int * group * count ;

fun ue1 (Elec (cand, _, _, _)) = cand ;

fun push x xsref = xsref := x :: !xsref ;

(* toList : 'a array -> 'a list *)
fun toList array = Array.foldr (op ::) [] array ;

(* numl : int -> 'a list -> (int * 'a) list
  number items in a list, starting from i *)
fun numl i [] = [] | numl i (x :: xs) = (i, x) :: numl (i+1) xs ;

(* modify a single array element by function f *)
fun updatef a i f = Array.update (a, i, f (Array.sub (a, i))) ;

datatype info = Info of {candStatus : status array, candTot : int array,
  ncont : int ref, (* number continuing, including those elected *)
  nelec : int ref, (* number elected *)
  count : count ref, (* number of current count *)
  fixed : {nc : int, (* number of candidates *)
	  nv : int, (* number of votes *) 
	  nq : int, (* quota *) 
	  ne : int (* number to be elected *) }, 
  results : (int * results) list ref,
  events : (int * elecEvent) list ref,
  (* Commissioner's lot, result expressed in terms of position in list,
    counting from 0 *)
  commLot : count -> cand list -> int} ;

val inforef = ref (Info {candStatus = Array.fromList [],
  candTot = Array.fromList [],
  ncont = ref ~1, nelec = ref ~1, count = ref ~1,
  fixed = {nc = ~1, nq = ~1, ne = ~1, nv = ~1}, 
  results = ref [], events = ref [],
  commLot = fn _ => fn _ => 0}) : info ref ;

val print = TextIO.print ;

end ; (* structure Types *)
